part of dartemporal;

/// The Is TE provides funcationality to test for equality of the various DateTime components such as Year, Month, etc. as a Temporal Expression.
class TEIs extends TExpression {
  DateTime _first;  // The first DateTime that matches this TE
  DateTime _last;   // The last DateTime that matches this TE
  int _teType;      // Keeps track of the "type" of TE that we are tracking
  var _value;       // Holds the value provided ot the constructor

  /// Returns the first DateTime that can be matched by a particular TE, or null if none exists.  The first() method should used a cached value that is calculated at construction where possible.  This will reduce the calculation overhead for other methods.
  DateTime first() {
    return _first;
  }

  /// Returns the last DateTime that can be matched by a particular TE, or null if none exists.  The last() method should used a cached value that is calculated at construction where possible.  This will reduce the calculation overhead for other methods.
  DateTime last() {
    return _last;
  }

  /// Allow this class to be "stringified", this does not return the JSON string but does return a Map that can be stringified.
  Map toJson() {
    if(_value is int) {
      return ({"TEMasterType": "TEIs", "TEType": _teType, "Value": _value});
    } else if(_value is DateTime) {
      return({"TEMasterType": "TEIs", "TEType": _teType, "Value": {"TEMasterType":"DateTime", "Value": _value.millisecondsSinceEpoch}});
    }
  }

  /// Create a TEIs object from a Map object that was created manually (why?) or from the toJson method
  ///   is a class factory
  factory TEIs.fromJson(Map source) {
    var value = source["Value"];
    TEIs result;

    if(value is Map) {
      value = new DateTime.fromMillisecondsSinceEpoch(value["Value"], isUtc: true);
    }

    switch(source["TEType"]) {
      case TExpression.YEAR:
        result = new TEIs.Year(value);
        break;
      case TExpression.MONTH:
        result = new TEIs.Month(value);
        break;
      case TExpression.DATETIME:
        result = new TEIs.DateTime(value);
        break;
      case TExpression.DAYOFWEEK:
        result = new TEIs.DayOfWeek(value);
        break;
      case TExpression.DAYOFMONTH:
        result = new TEIs.DayOfMonth(value);
        break;
      case TExpression.WEEKOFYEAR:
        result = new TEIs.WeekOfYear(value);
        break;
      case TExpression.DAYOFYEAR:
        result = new TEIs.DayOfYear(value);
        break;
      case TExpression.HOUR:
        result = new TEIs.Hour(value);
        break;
      case TExpression.MINUTE:
        result = new TEIs.Minute(value);
        break;
      case TExpression.SECOND:
        result = new TEIs.Second(value);
        break;
      case TExpression.MILLISECOND:
        result = new TEIs.Millisecond(value);
        break;
      case TExpression.DATE:
        result = new TEIs.Date(value);
        break;
      case TExpression.WEEKOFMONTH:
        result = new TEIs.WeekOfMonth(value);
        break;
    }
    return result;
  }

  /// Constructor to match all DateTimes in a specifically supplied year. Because we have a specific year, first() and last() will be able to return DateTimes, and not the null familiar to many other methods.
  TEIs.Year(int year) {
    _first = new DateTime.utc(year,1,1,0,0,0,0);
    _last = new DateTime.utc(year + 1,1,1,0,0,0,-1);
    _teType = TExpression.YEAR;
    _value = year;
  }

  /// Constructor to match all DateTimes that occur in a given month, regardless of year.
  TEIs.Month(int month) {
    _first = null;        // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.MONTH;
    _value = month;
  }

  /// Constructor to match all DateTimes that occur on a specific day of the week (Monday for example) regardless of year
  /// Uses the DateTime constants (DateTime.MONDAY) to select
  TEIs.DayOfWeek(int day) {
    _first = null;       // first and last are null because year, month, and week are undefined
    _last  = null;
    _teType = TExpression.DAYOFWEEK;
    _value = day;
  }

  /// Constructor to match all DateTimes that occur on a specific day of every month (3rd of every month for example) regardless of year
  TEIs.DayOfMonth(int day) {
    _first = null;      // first and last are null because year and month are undefined
    _last = null;
    _teType = TExpression.DAYOFMONTH;
    _value = day;
  }

  /// Constructor to match all DateTimes that occur on a specific day of every year (day 212 of every year)
  TEIs.DayOfYear(int day) {
    _first = null;     // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.DAYOFYEAR;
    _value = day;
  }

  /// Constructor to match all DateTimes that occur during a specific hour of every day.
  TEIs.Hour(int hour) {
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.HOUR;
    _value = hour;
  }

  /// Constructor to match all DateTimes that occur during a specific minute of every hour;
  TEIs.Minute(int minute) {
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.MINUTE;
    _value = minute;
  }

  /// Constructor to match all DateTimes that occur during a specific second of every minute;
  TEIs.Second(int second) {
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.SECOND;
    _value = second;
  }

  /// Constructor to match all DateTimes that occur during a specific millisecond of every second;
  TEIs.Millisecond(int millisecond) {
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.MILLISECOND;
    _value = millisecond;
  }

  /// Constructor to match all dateTimes that occur during a specific week of the year.
  TEIs.WeekOfYear(int week) {
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.WEEKOFYEAR;
    _value = week;
  }

  /// Constructor to match all DateTimes that occur exactly at given DateTime
  TEIs.DateTime(DateTime date) {
    _first = date;
    _last = date;
    _teType = TExpression.DATETIME;
    _value = date;
  }

  /// Constructor to match all DateTimes that occur on a specific day (ignoring supplied time parameters provided (if any) supplied in the DateTime
  TEIs.Date(DateTime date) {
    _first = new DateTime.utc(date.year, date.month, date.day, 0, 0, 0, 0);
    _last = new DateTime.utc(date.year, date.month, date.day + 1, 0, 0, 0, -1);
    _teType = TExpression.DATE;
    _value = date;
  }

  /// Constructor to match the week of a month.
  /// Positive values indicate a count starting from the beginning of the month
  /// Negative values indicate a count starting from the end of the month.
  /// A value of 0 will return false
  TEIs.WeekOfMonth(int week) {
    _first = null;
    _last = null;
    _teType = TExpression.WEEKOFMONTH;
    _value = week;
  }

  /// Test to see if an arbitrary testDate is a match for this TE
  bool includes(DateTime testDate) {
    bool result = false;
    switch(_teType) {
      case TExpression.YEAR:
        result = testDate.year == _value;
        break;
      case TExpression.MONTH:
        result = testDate.month == _value;
        break;
      case TExpression.DATETIME:
        result = testDate.isAtSameMomentAs(_value);
        break;
      case TExpression.DAYOFWEEK:
        result = testDate.weekday == _value;
        break;
      case TExpression.DAYOFMONTH:
        result = testDate.day == _value;
        break;
      case TExpression.WEEKOFYEAR:
        result = weekOfYear(testDate) == _value;
        break;
      case TExpression.DAYOFYEAR:
        result = dayOfYear(testDate) == _value;
        break;
      case TExpression.HOUR:
        result = testDate.hour == _value;
        break;
      case TExpression.MINUTE:
        result = testDate.minute == _value;
        break;
      case TExpression.SECOND:
        result = testDate.second == _value;
        break;
      case TExpression.MILLISECOND:
        result = testDate.millisecond == _value;
        break;
      case TExpression.DATE:
        result = (testDate.millisecondsSinceEpoch >= _first.millisecondsSinceEpoch) && (testDate.millisecondsSinceEpoch <= _last.millisecondsSinceEpoch);
        break;
      case TExpression.WEEKOFMONTH:
        result = false;
        int testWOM = weekOfMonth(testDate);

        if(_value >= 0) {
          result = _value == testWOM;
        } else {
          int maxWOM = maxWeekOfMonth(testDate);
          result = ((maxWOM + _value) + 1) == testWOM;
        }
        break;
    }
    return result;
  }

}