part of dartemporal;

/// The Is TE provides funcationality to test for occurrences before a deadline of the various DateTime components such as Year, Month, etc.
class TEIsBefore extends TExpression {
  DateTime _first;  // The first DateTime that matches this TE
  DateTime _last;   // The last DateTime that matches this TE (inclusive)
  int _teType;      // Keeps track of the "type" of TE that we are tracking
  bool _inclusive;  // Is true if this range is inclusive, false otherwise
  var _value;       // Holds the value used for testing
  var _paramValue;  // Holds the value provided in the constructor

  /// Returns the DateTime on which this TE starts matching
  DateTime first() {
    return _first;
  }

  /// Returns the DateTime on which this TE stops matching
  DateTime last() {
    return _last;
  }

  /// Allow this class to be "stringified", this does not return the JSON string but does return a Map that can be stringified.
  Map toJson() {
    if(_value is int) {
      return ({"TEMasterType": "TEIsBefore", "TEType": _teType, "Inclusive": _inclusive, "Value": _paramValue});
    } else if(_value is DateTime) {
      return({"TEMasterType": "TEIsBefore", "TEType": _teType, "Inclusive": _inclusive, "Value": {"TEMasterType":"DateTime", "Value": _paramValue.millisecondsSinceEpoch}});
    }
  }

  /// Create a TEIs object from a Map object that was created manually (why?) or from the toJson method
  ///   is a class factory
  factory TEIsBefore.fromJson(Map source) {
    var value = source["Value"];
    bool inclusive = source["Inclusive"];
    TEIsBefore result;

    if(value is Map) {
      value = new DateTime.fromMillisecondsSinceEpoch(value["Value"], isUtc: true);
    }

    switch(source["TEType"]) {
      case TExpression.YEAR:
        result = new TEIsBefore.Year(value,inclusive);
        break;
      case TExpression.MONTH:
        result = new TEIsBefore.Month(value,inclusive);
        break;
      case TExpression.DATETIME:
        result = new TEIsBefore.DateTime(value,inclusive);
        break;
      case TExpression.DAYOFWEEK:
        result = new TEIsBefore.DayOfWeek(value,inclusive);
        break;
      case TExpression.DAYOFMONTH:
        result = new TEIsBefore.DayOfMonth(value,inclusive);
        break;
      case TExpression.WEEKOFYEAR:
        result = new TEIsBefore.WeekOfYear(value,inclusive);
        break;
      case TExpression.DAYOFYEAR:
        result = new TEIsBefore.DayOfYear(value,inclusive);
        break;
      case TExpression.HOUR:
        result = new TEIsBefore.Hour(value,inclusive);
        break;
      case TExpression.MINUTE:
        result = new TEIsBefore.Minute(value,inclusive);
        break;
      case TExpression.SECOND:
        result = new TEIsBefore.Second(value,inclusive);
        break;
      case TExpression.MILLISECOND:
        result = new TEIsBefore.Millisecond(value,inclusive);
        break;
      case TExpression.DATE:
        result = new TEIsBefore.Date(value,inclusive);
        break;
      case TExpression.WEEKOFMONTH:
        result = new TEIsBefore.WeekOfMonth(value);
        break;
    }
    return result;
  }

  /// Constructor to match all DateTimes before given year.
  /// By default this is inclusive, but that can be changed with the second boolean parameter.
  TEIsBefore.Year(int year, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;
    _value = year;
    _paramValue = year;
    if(inclusive == true) { _value += 1; };   // All tests are non inclusive so if it needs to be inclusive add one
    _last = new DateTime.utc(_value, 1, 1, 0, 0, 0, -1);
    _teType = TExpression.YEAR;
  }

  /// Constructor to match all DateTimes that occur in a given month, regardless of year.
  TEIsBefore.Month(int month, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;        // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.MONTH;
    _value = month;
    _paramValue = month;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur on a specific day of the week (Monday for example) regardless of year
  /// Uses the DateTime constants (DateTime.MONDAY) to select
  TEIsBefore.DayOfWeek(int day, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;       // first and last are null because year, month, and week are undefined
    _last  = null;
    _teType = TExpression.DAYOFWEEK;
    _value = day;
    _paramValue = day;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur on a specific day of every month (3rd of every month for example) regardless of year
  TEIsBefore.DayOfMonth(int day, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;      // first and last are null because year and month are undefined
    _last = null;
    _teType = TExpression.DAYOFMONTH;
    _value = day;
    _paramValue = day;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur on a specific day of every year (day 212 of every year)
  TEIsBefore.DayOfYear(int day, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;     // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.DAYOFYEAR;
    _value = day;
    _paramValue = day;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur during a specific hour of every day.
  TEIsBefore.Hour(int hour, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.HOUR;
    _value = hour;
    _paramValue = hour;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur during a specific minute of every hour;
  TEIsBefore.Minute(int minute, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.MINUTE;
    _value = minute;
    _paramValue = minute;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur during a specific second of every minute;
  TEIsBefore.Second(int second, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.SECOND;
    _value = second;
    _paramValue = second;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur during a specific millisecond of every second;
  TEIsBefore.Millisecond(int millisecond, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.MILLISECOND;
    _value = millisecond;
    _paramValue = millisecond;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all dateTimes that occur during a specific week of the year.
  TEIsBefore.WeekOfYear(int week, [bool inclusive = true]) {
    _inclusive = inclusive;
    _first = null;    // first and last are null because year is undefined
    _last = null;
    _teType = TExpression.WEEKOFYEAR;
    _value = week;
    _paramValue = week;
    if(inclusive == true) { _value += 1; };  // All tests are non inclusive so if it needs to be inclusive add one
  }

  /// Constructor to match all DateTimes that occur before a given Date
  TEIsBefore.DateTime(DateTime date, [bool inclusive = true]) {
    _first = null;
    _last = date;
    _inclusive = inclusive;
    _teType = TExpression.DATETIME;
    _value = date;
    _paramValue = date;
    if(inclusive == true) {
      _value = _value.add(new Duration(milliseconds:1));
    } else {
      _last = _last.subtract(new Duration(milliseconds:1));
    }
  }

  /// Constructor to match all DateTimes that occur after a specific day (ignoring supplied time parameters provided (if any) supplied in the DateTime
  TEIsBefore.Date(DateTime date, [bool inclusive = true]) {
    _first = null;
    _last = new DateTime.utc(date.year, date.month, date.day + 1,0,0,0,-1);
    _inclusive = inclusive;
    _teType = TExpression.DATE;
    _value = date;
    _paramValue = date;
    if(inclusive == true) {
      _value = _value.add(new Duration(milliseconds:1));
    } else {
      _last = _last.subtract(new Duration(hours:24));
    }
  }


  /// Constructor to match all weeks before a given week of a month.
  /// Positive values indicate a count starting from the beginning of the month
  /// Negative values indicate a count starting from the end of the month.
  /// This TI is always inclusive;
  /// A value of 0 will return false
  TEIsBefore.WeekOfMonth(int week) {
    _first = null;
    _last = null;
    _inclusive = true;
    _teType = TExpression.WEEKOFMONTH;
    _value = week;
    _paramValue = week;
  }

  /// Test to see if an arbitrary testDate is a match for this TE
  bool includes(DateTime testDate) {
    bool result = false;
    switch(_teType) {
      case TExpression.YEAR:
        result = testDate.year < _value;
        break;
      case TExpression.MONTH:
        result = testDate.month < _value;
        break;
      case TExpression.DATETIME:
        result = testDate.isBefore(_value);
        break;
      case TExpression.DAYOFWEEK:
        result = testDate.weekday < _value;
        break;
      case TExpression.WEEKOFYEAR:
        result = weekOfYear(testDate) < _value;
        break;
      case TExpression.DAYOFMONTH:
        result = testDate.day < _value;
        break;
      case TExpression.DAYOFYEAR:
        result = dayOfYear(testDate) < _value;
        break;
      case TExpression.HOUR:
        result = testDate.hour < _value;
        break;
      case TExpression.MINUTE:
        result = testDate.minute < _value;
        break;
      case TExpression.SECOND:
        result = testDate.second < _value;
        break;
      case TExpression.MILLISECOND:
        result = testDate.millisecond < _value;
        break;
      case TExpression.DATE:
        result = testDate.millisecondsSinceEpoch < _value.millisecondsSinceEpoch;
        break;
      case TExpression.WEEKOFMONTH:
        result = false;
        int testWOM = weekOfMonth(testDate);
        if(_value == 0) {
          result = false;
        } else if(_value >= 0) {
          result = _value >= testWOM;
        } else {
          int maxWOM = maxWeekOfMonth(testDate);
          result = ((maxWOM + _value) + 1) >= testWOM;
        }
        break;
    }
    return result;
  }

}